/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.proofpoint.galaxy.coordinator;

import com.google.common.base.Function;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.inject.Inject;
import com.proofpoint.galaxy.shared.AgentStatus;
import com.proofpoint.galaxy.shared.Assignment;
import com.proofpoint.galaxy.shared.AssignmentRepresentation;
import com.proofpoint.galaxy.shared.SlotStatus;
import com.proofpoint.galaxy.shared.SlotStatusRepresentation;

import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.util.List;
import java.util.UUID;

import static com.google.common.collect.Collections2.transform;
import static com.proofpoint.galaxy.coordinator.StringFunctions.toStringFunction;
import static com.proofpoint.galaxy.shared.SlotStatusRepresentation.fromSlotStatusWithShortIdPrefixSize;
import static java.lang.Math.max;

@Path("/v1/slot")
public class CoordinatorSlotResource
{
    private final Coordinator coordinator;

    public static final int MIN_PREFIX_SIZE = 4;

    @Inject
    public CoordinatorSlotResource(Coordinator coordinator)
    {
        Preconditions.checkNotNull(coordinator, "coordinator must not be null");

        this.coordinator = coordinator;
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    public Response getAllSlots(@Context UriInfo uriInfo)
    {
        List<UUID> uuids = Lists.transform(coordinator.getAllSlotStatus(), SlotStatus.uuidGetter());

        int prefixSize = MIN_PREFIX_SIZE;
        if (!uuids.isEmpty()) {
            prefixSize = max(prefixSize, Strings.shortestUniquePrefix(transform(uuids, toStringFunction())));
        }

        Predicate<SlotStatus> slotFilter = SlotFilterBuilder.build(uriInfo, false, uuids);

        Iterable<SlotStatusWithExpectedState> result = coordinator.getAllSlotStatusWithExpectedState(slotFilter);

        return Response.ok(Iterables.transform(result, toSlotRepresentation(prefixSize))).build();
    }

    public static Function<SlotStatusWithExpectedState, SlotStatusRepresentation> toSlotRepresentation(final int size)
    {
        return new Function<SlotStatusWithExpectedState, SlotStatusRepresentation>()
        {
            public SlotStatusRepresentation apply(SlotStatusWithExpectedState status)
            {
                return from(status, size);
            }
        };
    }

    public static SlotStatusRepresentation from(SlotStatusWithExpectedState slotStatusWithExpectedState, int shortIdPrefixSize)
    {
        SlotStatus slotStatus = slotStatusWithExpectedState.getSlotStatus();
        ExpectedSlotStatus expectedSlotStatus = slotStatusWithExpectedState.getExpectedSlotStatus();

        String expectedBinary = null;
        String expectedConfig = null;
        String expectedStatus = null;
        if (expectedSlotStatus != null) {
            expectedBinary = expectedSlotStatus.getBinary();
            expectedConfig = expectedSlotStatus.getConfig();
            expectedStatus = expectedSlotStatus.getStatus() == null ? null : expectedSlotStatus.getStatus().toString();
        }

        if (slotStatus.getAssignment() != null) {
            return new SlotStatusRepresentation(slotStatus.getId(),
                    slotStatus.getId().toString().substring(0, shortIdPrefixSize),
                    slotStatus.getName(),
                    slotStatus.getSelf(),
                    slotStatus.getLocation(),
                    slotStatus.getAssignment().getBinary().toString(),
                    slotStatus.getAssignment().getConfig().toString(),
                    slotStatus.getState().toString(),
                    slotStatus.getStatusMessage(),
                    slotStatus.getInstallPath(),
                    expectedBinary,
                    expectedConfig,
                    expectedStatus
            );
        }
        else {
            return new SlotStatusRepresentation(slotStatus.getId(),
                    slotStatus.getId().toString().substring(0, shortIdPrefixSize),
                    slotStatus.getName(),
                    slotStatus.getSelf(),
                    slotStatus.getLocation(),
                    null,
                    null,
                    slotStatus.getState().toString(),
                    slotStatus.getStatusMessage(),
                    slotStatus.getInstallPath(),
                    expectedBinary,
                    expectedConfig,
                    expectedStatus
            );

        }
    }

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public Response install(
            AssignmentRepresentation assignmentRepresentation,
            @DefaultValue("1") @QueryParam("limit") int limit,
            @Context UriInfo uriInfo)
    {
        Preconditions.checkNotNull(assignmentRepresentation, "assignmentRepresentation must not be null");
        Preconditions.checkArgument(limit > 0, "limit must be at least 1");

        Assignment assignment = assignmentRepresentation.toAssignment();

        Predicate<AgentStatus> agentFilter = AgentFilterBuilder.build(uriInfo);
        List<SlotStatus> results = coordinator.install(agentFilter, limit, assignment);

        List<UUID> uuids = Lists.transform(coordinator.getAllSlotStatus(), SlotStatus.uuidGetter());
        int uniquePrefixSize = max(Strings.shortestUniquePrefix(transform(uuids, toStringFunction())), MIN_PREFIX_SIZE);

        return Response.ok(transform(results, fromSlotStatusWithShortIdPrefixSize(uniquePrefixSize))).build();
    }

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    public Response terminateSlots(@Context UriInfo uriInfo)
    {
        List<UUID> uuids = Lists.transform(coordinator.getAllSlotStatus(), SlotStatus.uuidGetter());

        int prefixSize = MIN_PREFIX_SIZE;
        if (!uuids.isEmpty()) {
            prefixSize = max(prefixSize, Strings.shortestUniquePrefix(transform(uuids, toStringFunction())));
        }

        Predicate<SlotStatus> slotFilter = SlotFilterBuilder.build(uriInfo, true, uuids);

        List<SlotStatus> result = coordinator.terminate(slotFilter);

        return Response.ok(transform(result, fromSlotStatusWithShortIdPrefixSize(prefixSize))).build();
    }
}
